<?php

/*
    Web Application Deployment System
    (c)2006-2007 Tim Jackson (tim@timj.co.uk)
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of version 3 of the GNU General Public License as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

define('WADF_VCREFTYPE_REV','rev');
define('WADF_VCREFTYPE_TAG','tag');
define('WADF_VCREFTYPE_BRANCH','branch');
define('WADF_VCREFTYPE_TRUNK','trunk');

/**
 * Main WADF class
 */
class Tools_WADF {

	var $_debug = 0;

	// the app reference
	var $appref;

	// macro definitions
	var $_macro_defs;
	
	// resolved macro values
	var $_macro_values = array();
	
	// generic class options
	var $_options = array(
		'master_config' => '@data_dir@/Tools_WADF/wadf.conf',
		'webserver_type' => 'httpd',
		'update_hosts' => false,
	);
	
	// macros which we don't manage to resolve during deployment
	var $_unresolved_macros = array();
	
	function Tools_WADF ($appref, $options=null, $cmdline_macros=null)
	{
		// Strip trailing slash from appref, if present
		if (substr($appref, -1, 1) == '/') {
			$appref = substr($appref, 0, strlen($appref)-1);
		}
		
		$this->_options['confd_dir'] = $_ENV['HOME'].'/.wadf/wadf.d';
	
		$valid_options = array('master_config');
		if (is_array($options)) {
			foreach ($options as $option => $value) {
				if (in_array($option, $valid_options)) {
					$this->_options[$option] = $value;
				} else {
					trigger_error("Invalid config option '$option' passed to constructor",E_USER_ERROR);
				}
			}
		}
		
		$this->_setInternalMacros();
		
		// set default options
		$this->_appendMacroDefs($this->_options);
		
		// set appref
		$this->appref = $appref;
		$options['appref'] = $appref;
		
		// set user-supplied options
		$this->_appendMacroDefs($options);

		if (!is_array($cmdline_macros)) {
			$cmdline_macros = array();
		}
		$this->processConfigs($cmdline_macros);
		
		$this->resolveAllMacros();
	}
	
	function _setInternalMacros()
	{
		if (isset($_ENV['HOSTNAME'])) {
			$macros['hostname'] = $_ENV['HOSTNAME'];
		} else {
			$macros['hostname'] = gethostbyaddr('127.0.0.1'); // FIXME this can't be the right way of doing it
		}
		$macros['cwd'] = getcwd();
		if (function_exists('posix_getuid')) {
			$details = posix_getpwuid(posix_getuid());
			$macros['user'] = $details['name'];
		} else {
			$macros['user'] = 'UNKNOWN';
		}
		if (isset($_ENV['HOME'])) {
			$macros['home'] = $_ENV['HOME'];
		}
		$this->_appendMacroDefs($macros);
	}
	
	// branch = [a branch], trunk, tags
	// rev = [rev no], HEAD
	function deploy($rev, $revtype=WADF_VCREFTYPE_TRUNK, $db_deploy=null,$db_deploy_pass='')
	{		
		$dir = $this->resolveMacro('deploy_path');
		$this->checkout($dir, $rev, $revtype);
		$this->setupPEAR($dir);
		$this->processTemplatesInDir($dir);
		
		// TODO: handled unresolved macros? or just let UI do it?
		
		$this->deployDatabase($db_deploy_pass);
		$this->deployVhost();
		$this->deployDNS();
		$this->runKickstart();
	}
	
	function getUnresolvedMacros()
	{
		return $this->_unresolved_macros;
	}

	function unDeploy()
	{
		$this->unDeployDatabase();
		// TODO remove files, remove vhost, remove DNS
	}
	
	function _enumerateMultipleEntities()
	{
		$dir = $this->resolveMacro('deploy_path');
	
		$macros_in_templates = $this->extractMacrosFromDir($dir);
		// we need to enumerate the db and vhost numbers
		$in_use = array();
		foreach ($macros_in_templates as $macro) {
			if (preg_match('/^(db|vhost)(\d+)_/', $macro, $matches)) {
				// array like
				// in_use = array (
				//   db => array (1=>true, 2=>true, 3=>true)
				//   vhost => array (1=>true, 2=>true)
				$in_use[$matches[1]][$matches[2]] = true;
			}
		}
		// now turn the array into db => array(1,2,3) etc.
		foreach ($in_use as $type => $list_of_valid) {
			$in_use[$type] = array_keys($list_of_valid);
		}
		
		return $in_use;
	}
	
	function deployDatabase($rootpass='')
	{
		$in_use = $this->_enumerateMultipleEntities();
		$dir = $this->resolveMacro('deploy_path');
		
		// deploy database
		if (isset($in_use['db'])) {
			foreach ($in_use['db'] as $num) {
				$host = $this->resolveMacro("db${num}_host");
				$user = $this->resolveMacro("db${num}_user");
				$pass = $this->resolveMacro("db${num}_pass");
				$name = $this->resolveMacro("db${num}_name");
				$type = $this->resolveMacro("db${num}_type");
				$schema = $this->resolveMacro("db${num}_schema");
				if ($type != 'mysql') {
					die ("Unsupported database type '$type'");
				}
				// TODO check host, user, pass, name are all non-empty
				print "Setting up database $name on host $host...\n";
				$schema_path = "$dir/$schema";
				if (file_exists($schema_path)) {
					// FIXME abstract db-specific stuff
					// FIXME error handling!
					// FIXME quote strings!
					$db = @mysql_connect($host, 'root', $rootpass);
					if (!is_resource($db)) die("Could not connect to database: ".mysql_error()."\n");
					mysql_query("CREATE DATABASE $name", $db);
					mysql_query("GRANT ALL on $name.* to $user IDENTIFIED BY '$pass'", $db);
					mysql_close($db);
					$cmd = "mysql $name -h $host -u $user ".(($pass != null) ? "-p$pass" : '')." < $dir/$schema";
					//print "$cmd\n";
					exec($cmd);
				} else {
					die("ERROR: schema file $schema_path not found\n");
				}
			}
		} else {
			print "No database to deploy\n";
		}
		
	}
	
	function unDeployDatabase($rootpass='')
	{
		$in_use = $this->_enumerateMultipleEntities();
		if (isset($in_use['db'])) {
			foreach ($in_use['db'] as $num) {
				$host = $this->resolveMacro("db${num}_host");
				$name = $this->resolveMacro("db${num}_name");
				// TODO error handling
				$db = @mysql_connect($host, 'root', $rootpass);
				mysql_query("DROP DATABASE $name");
				mysql_close($db);
				// TODO remove DB users? But what if other site uses them?
			}
		}
	}
	
	function deployDNS()
	{
		// TODO other ways of deploying DNS
		$deploy_type = $this->resolveMacro('deploy_dns');

		if ($deploy_type == 'none') {
			return;
		}

		$in_use = $this->_enumerateMultipleEntities();
		if (isset($in_use['vhost'])) {
			foreach ($in_use['vhost'] as $vhost_id) {
				$hosts[] = $this->resolveMacro("vhost${vhost_id}_name");
			}

			switch ($deploy_type) {
				case 'hosts':
					$this->addHostsToHostsFile($hosts);
					break;
				default:
					trigger_error("Unknown DNS deployment method",E_USER_ERROR);
					break;
			}
		}
	}
	
	function addHostsToHostsFile($hosts)
	{
		$hosts_file = $this->resolveMacro('deploy_dns_hosts_file');
		$deploy_ip = $this->resolveMacro('deploy_dns_hosts_ip');
		$existing_hostnames = array();
		$file_contents = file($hosts_file);
		$line_num = 0;
		$localhost_line_num = 0;
		$base_host = ''; //the first host found
		foreach ($file_contents as $line) {
			if (substr($line, 0, strlen($deploy_ip)) == $deploy_ip) {
				if (preg_match_all('/[\t\s]+(\S+)/', $line, $matches)) {
					foreach ($matches[1] as $host) {
						$existing_hostnames[] = trim($host);
						if (!$base_host) $base_host = trim($host);
					}
				}
				$localhost_line_num = $line_num;
			}
			$line_num++;
		}
		foreach ($hosts as $host) {
			if (!in_array($host, $existing_hostnames)) {
				array_splice($file_contents, $localhost_line_num + 1, 0, "$deploy_ip\t$base_host\t$host\n");
			}
		}
		$fp = @fopen($hosts_file, 'w');
		if (is_resource($fp)) {
			fputs($fp, implode('', $file_contents));
			fclose($fp);
		} else {
			print "WARNING: Could not update list of local hostnames in hosts file '$hosts_file'. Check permissions.\n";
		}
	}
	
	function deployVhost($dir=null)
	{
		if ($dir === null) {
			$dir = $this->resolveMacro('deploy_path');
		}

		// deploy vhost config
		$source_file = $dir.'/'.$this->resolveMacro('vhost_config_template');
		$dest_file = $this->resolveMacro('vhost_config_path').'/'.$this->resolveMacro('instance').'.conf';
		//print "Copying $source_file to $dest_file\n";
		
		$config = file_get_contents($source_file);
		
		// If we are using mod_php, insert the PHP config options
		if ($this->resolveMacro('php_type') == 'mod_php') {
			$php_ini = $dir . '/' . $this->resolveMacro('php_config_location');
			$directives = $this->_convertPhpIniToApacheModPhp($php_ini);
			
			// Stick the PHP directives at the end of the vhost
			$php_directives = implode("\n\t", $directives);
			$config = str_replace("</VirtualHost>", "\n\t# PHP directives processed from $php_ini\n\t" . $php_directives . "\n" . '</VirtualHost>', $config);
			
			// Process local PHP override ini file, if it exists
			$php_ini_local = $this->resolveMacro('php_config_location_extra');
			if ($php_ini_local != '@php_config_location_extra@') {
				$directives_local = $this->_convertPhpIniToApacheModPhp($php_ini_local);
				if (count($directives_local) > 0) {
					$php_directives = implode("\n\t", $directives_local);
					$config = str_replace("</VirtualHost>", "\n\t# PHP directives processed from $php_ini_local\n\t" . $php_directives . "\n" . '</VirtualHost>', $config);
				}
			}
		} else {
			$this->_deployPHPConfig($dir);
		}
		
		$config = "# wadf-working-copy: $dir\n$config";
		$fp = fopen($dest_file, 'w');
		if (is_resource($fp)) {
			fputs($fp, $config);
			fclose($fp);
		} else {
			print "Could not open vhost config destination file '$dest_file'\n";
		}
		
		// restart apache
		$restart_cmd = $this->resolveMacro('webserver_restart_cmd');
		print "Restarting webserver...\n";
		if ($restart_cmd != '@webserver_restart_cmd@') {
			passthru($restart_cmd);
		}
	}
	
	/**
	 * @param string $file Filename to read. If file does not exist, no error will be thrown.
	 * @return array Array of Apache mod_php config rules
	 */
	function _convertPhpIniToApacheModPhp($file)
	{
		$directives_out = array();
		if (file_exists($file)) {
			print "Processing PHP config file $file...\n";
			$php_ini_directives = file($file);
			$directives_out = array();
			foreach ($php_ini_directives as $directive) {
				if (preg_match('/^\s*;/', $directive)) {
					$directives_out[] = $directive;
				} else if (preg_match('/^([a-z0-9_]+)\s*=\s*(.+)$/', $directive, $matches)) {
					$directive_name = trim($matches[1]);
					$directive_value = trim($matches[2]);
					
					// See if it looks like a flag or a value
					if (in_array(strtolower($directive_value), array('0','1','on','off'))) {
						$directives_out[] = "php_admin_flag $directive_name $directive_value";
					} else {
						$directives_out[] = "php_admin_value $directive_name $directive_value";
					}
				} else {
					print "WARNING: Could not parse PHP ini config line from $file:\n\t$directive\n";
				}
			}
		}
		return $directives_out;
	}
	
	// Not needed for mod_php; it is done in deployVhost();
	function _deployPHPConfig($dir)
	{
		$php_type = $this->resolveMacro('php_type');
		if ($php_type == 'mod_php') {
			return; // do nothing
		} else if (preg_match('/^cgi:(.+)$/', $php_type, $matches)) {
			$php_ini_dest = trim($matches[1]);
			$php_ini_source = $dir . '/' . $this->resolveMacro('php_config_location');
			if (file_exists($php_ini_source)) {
				print "Deploying PHP config file $php_ini...\n";
				copy($php_ini_source, $php_ini_dest);
			}
		} else {
			print "ERROR: Unknown PHP type '$php_type'\n";
		}
	}
	
	// FIXME abstract
	function checkout ($destdir, $rev, $revtype=WADF_VCREFTYPE_TRUNK)
	{
		// MODULE is appref
		// BASE URL is vc_base
		
		switch ($revtype) {
			case WADF_VCREFTYPE_REV:
				$svn_path = '';
				$options = "-r $rev";
				break;
			case WADF_VCREFTYPE_TRUNK:
				$svn_path = 'trunk';
				$options = "-r $rev";
				break;
			case WADF_VCREFTYPE_BRANCH:
				$svn_path = "branches/$rev";
				break;
			case WADF_VCREFTYPE_TAG:
				$svn_path = "tags/$rev";
				break;
		}
		
		$vc_base = $this->resolveMacro('vc_base');
		
		$cmd = "svn $options co $vc_base/$this->appref/$svn_path $destdir";
		print "Running $cmd\n";
		exec($cmd, $output, $ret);
		print implode("\n",$output)."\n";
		if ($ret != 0) {
			print "Error checking out\n";
			exit(5);
		}
		
		if ($fp = @fopen("$destdir/.wadf-instance",'w')) {
			fputs($fp, $this->appref);
			fclose($fp);
		}
		
	}
	
	function setupPEAR($dir)
	{
		// See if we actually need PEAR for this deployment
		if (!file_exists("$dir/package.xml")) {
			return;
		}
		
		$pear_path = $this->resolveMacro('dep_pear_deploy_path');
		$standalone_pear = false;
		// if empty, we want to use already-installed PEAR
		if (!empty($pear_path)) {
			$standalone_pear = true;
			if (!file_exists("$pear_path/.pearrc")) {
				$cmd = "pear -v -c $pear_path/.pearrc ".
					"-d php_dir=$pear_path ".
					"-d data_dir=$pear_path/data ".
					"-d ext_dir=$pear_path/ext ".
					"-d doc_dir=$pear_path/docs ".
					"-d test_dir=$pear_path/tests ".
					"-d cache_dir=$pear_path/cache ".
					"-d download_dir=$pear_path/downloads ".
					"-d temp_dir=$pear_path/temp ".
					"-d bin_dir=$pear_path ".
					"-s";
				exec($cmd);
					
				print "Installing PEAR Core into $pear_path (config file is $pear_path/.pearrc)...\n";
				passthru("pear -c $pear_path/.pearrc install --onlyreqdeps pear.php.net/PEAR");
			}
			
			$base_channel = $this->resolveMacro('dep_pear_base_channel');
			if (!empty($base_channel)) {
				if (preg_match('/^(.+):(.+)@(.+)$/',$base_channel, $matches)) {
					$username = $matches[1];
					$password = $matches[2];
					$channel = $matches[3];
				} else {
					$channel = $base_channel;
				}
				if (empty($channel)) trigger_error("Channel empty", E_USER_ERROR);
				print "Discovering channel $channel...\n";
				
				// FIXME check if channel is already discovered and skip next step if so
				
				$this->_runPEAR("channel-discover $channel");
				if (isset($username)) {
					// FIXME interactive
	    			$this->_runPEAR("-d default_channel=$channel login");
				}
			}
		} else {
			exec($this->_getPEARCmd() . ' config-get php_dir', $output);
			$pear_path = $output[0];
			echo "Using existing PEAR installation in $pear_path\n";
		}
		
		$this->_installPEARBaseRoles();
		$this->_configurePEARBaseRoles($pear_path, $dir, $standalone_pear);
		
		print "Processing package.xml and installing PEAR dependencies...\n";
		$application_dir = '';
		if (!$standalone_pear) {
			// Deploy to the same directory. Is !$standalone_pear really the
			// right criteria to use here?
			$application_dir = "-d application_dir=$dir";
		}
		$this->_runPEAR("$application_dir install --onlyreqdeps -f $dir/package.xml");
		
		$this->_setPearMacros();
		
	}
	
	function _installPEARBaseRoles()
	{
	
		$base_roles = $this->resolveMacro('dep_pear_base_roles');
		
		if (!empty($base_roles)) {
			// The command line binary to run
			$pear = $this->_getPEARCmd();
			
			$roles = explode(' ', $base_roles);
			foreach ($roles as $package) {
				exec("$pear list $package", $out, $ret);
				if ($ret != 0) {
					print "Installing $package...\n";
					$this->_runPEAR("install --soft --onlyreqdeps $base_roles");
				}
			}
		}
	}
	
	function _configurePEARBaseRoles($pear_path, $dir, $standalone_pear=false)
	{
		$vars = array(
			'applicationcore_dir' => $pear_path . '/applications'
		);
		
		if ($standalone_pear) {
			$vars['application_dir'] = $dir;
		}
		
		$base_channel = $this->_getPEARBaseChannelName();
		
		// The command line binary to run
		$pear = $this->_getPEARCmd();
		
		// Set the variables if needed
		foreach ($vars as $var => $value) {
			exec("$pear config-get $var", $out, $ret);
			if (empty($out[0])) {
				print "Setting $var = $value\n";
				if (!empty($base_channel)) {
					$this->_runPEAR("config-set -c $base_channel $var \"$value\"");
				} else {
					$this->_runPEAR("config-set $var \"$value\"");
				}
			}
		}
	}
	
	/**
	 * Returns base channel name *without* any optional username/pass
	 */
	function _getPEARBaseChannelName()
	{
		$channel = null;
		$base_channel = $this->resolveMacro('dep_pear_base_channel');
		if (!empty($base_channel)) {
			if (preg_match('/^(.+):(.+)@(.+)$/',$base_channel, $matches)) {
				$channel = $matches[3];
			} else {
				$channel = $base_channel;
			}
			if (empty($channel)) trigger_error("Channel empty", E_USER_ERROR);
		}
		return $channel;
	}
	
	function _runPEAR($cmd)
	{
		$runcmd = $this->_getPEARCmd() . ' ' . $cmd;
		if ($this->_debug) print "Running: $runcmd\n";
		passthru($runcmd);
		return $ret;
	}
	
	function _getPEARCmd()
	{
		$pear_path = $this->resolveMacro('dep_pear_deploy_path');
		if (empty($pear_path)) {
			exec('pear config-get bin_dir', $output, $ret);
			$pear_path = $output[0];
			$runcmd = "$pear_path/pear";
		} else {
			$runcmd = "$pear_path/pear -c $pear_path/.pearrc";
		}
		return $runcmd;
	}
	
	/**
	 * Add a selection of PEAR config options to the macro list, prefixing each
	 * one with "pear_"
	 */
	function _setPearMacros()
	{
		$pear_macros = array('application_dir','applicationcore_dir','bin_dir','doc_dir','php_dir','data_dir');
		$macro_defs = array();
		foreach ($pear_macros as $macro) {
			$pear = $this->_getPEARCmd();
			unset($out);
			exec("$pear config-get $macro", $out);
			if (!empty($out[0])) {
				$macro_defs["pear_$macro"] = trim($out[0]);
			}
		}
		$this->_appendMacroDefs($macro_defs);
	}
	
	function processTemplatesInDir($dir)
	{
		$files = $this->_listTemplateFiles($dir);
		foreach ($files as $file) {
			$content = file_get_contents($file);
			$content = $this->resolveString($content, null, "file:$file");
			$output_file = preg_replace('/^(.+)\.template$/','\1',$file);
			$fout_fh = fopen($output_file, 'w+');
			fputs($fout_fh, $content);
			fclose($fout_fh);
		}
	}
	
	// go through all template files in $dir and find out which macros are used in them
	// returns an array of unique macros
	function extractMacrosFromDir($dir)
	{
		$macros = array();
		$files = $this->_listTemplateFiles($dir);
		foreach ($files as $file) {
			if ($this->_debug) $this->_debugOutput("extracting macros from file '$file'");
			$contents = file_get_contents($file);
			$new_macros = $this->extractMacrosFromString($contents);
			if ($this->_debug) $this->_debugOutput("macros in file '$file' = ".implode(',',$new_macros));
			$macros = array_merge($macros,$new_macros);
			unset($new_macros);
		}
		return $macros;
	}
	
	function _listTemplateFiles($dir)
	{
		exec("find $dir -name '*.template'", $output, $ret);
		return $output;
	}
	
	// takes a string like '@underscore:foo@@bar@' and returns array('underscore:foo','bar');
	// if keep_modifiers = false then returns array('foo','bar');
	function extractMacrosFromString($string, $keep_modifiers = true)
	{
		preg_match_all('/@(([a-z]+:)?([a-z0-9_]{3,30}))@/', $string, $matches, PREG_PATTERN_ORDER);
		
		$macros = array_unique($matches[1]);
		if ($keep_modifiers == false) {
			foreach ($macros as $macro) {
				$newmacros[] = $this->getBaseMacroName($macro);
			}
			$macros = $newmacros;
		}
		return $macros;
	}
	
	function runKickstart()
	{
		$kickstart_script = $this->resolveMacro('kickstart_script');
		if (!empty($kickstart_script)) {
			// Nasty hack to make sure the kickstart script is executable.
			// Really it should be installed with role=script and PEAR would
			// handle it, but then it would end up being installed in the wrong
			// place
			// We should really read the existing mode and just set the executable
			// bit, but that is really hard to do with PHP because chmod() doesn't
			// like a variable as the second parameter very much.
			chmod($kickstart_script, 0750);
			
			print "\nRunning kickstart script $kickstart_script...\n";
			
			// Run the script
			passthru($kickstart_script);
		}
	}

	// $options is an array of macros e.g. passed from command line
	function processConfigs($options=null)
	{
		/* get:
		- globals
		- per-user
		- profile
		- site-specific (count($this->_macro_values) > 0) {
			$this->_macro_values = array();
			$this->resolveAllMacros();
		}
		- command line
		in that order. lump all together linearly in above order
		then process line-by-line, with later ones taking precedence.
		*/
		
		if ($options === null) $options = array();
		
		// Read global config
		$this->readConfigFile($this->_options['master_config']);
		
		// Read per-user config
		if (isset($_ENV['HOME'])) {
			$wadfrc = $_ENV['HOME'].'/.wadf/config';
			if (file_exists($wadfrc)) {
				$this->readConfigFile($wadfrc);
			}
		}
		
		// Read conf.d
		$confd_dir = $this->_options['confd_dir'];
		print "Looking in $confd_dir...\n";
		if (file_exists("$confd_dir/$this->appref.conf")) {
			print "Using $confd_dir/$this->appref.conf\n";
			$this->readConfigFile("$confd_dir/$this->appref.conf");
		}
		
		// Append options from command line
		$this->_appendMacroDefs($options);
	}
	
	function _appendMacroDefsTop($array)
	{
		$copy = $this->_macro_defs;
		$this->_macro_defs = array();
		foreach ($array as $key => $value) {
			$this->_macro_defs[] = new Tools_WADF_MacroDef($key,$value);
		}
		foreach ($copy as $value) {
			$this->_macro_defs[] = $value;
		}
		$this->_rebuildMacroValues();
	}
	
	function _appendMacroDefs($array)
	{
		if (!is_array($array)) return;
		
		foreach ($array as $key => $value) {
			$this->_macro_defs[] = new Tools_WADF_MacroDef($key,$value);
		}
		$this->_rebuildMacroValues();
	}
	
	function _rebuildMacroValues()
	{
		if (count($this->_macro_values) > 0) {
			$this->resolveAllMacros();
		}
	}
	
	function readConfigFile($file)
	{
		$config = parse_ini_file($file, TRUE);
		if (isset($config['globals'])) {
			print "Setting global configuration options from file $file\n";
			$this->_appendMacroDefs($config['globals']);
		}
		// We need to do this to ensure "profile" is set.
		// Possible optimisation: treat "profile" specially, as an unexpandable
		// macro, and create a method something like "getLatestMacro" which gets
		// the most recent definition of a given "macro", and use the literal
		// contents of that.
		$this->resolveAllMacros();
		$profile = $this->resolveMacro('profile');
		if ($profile == '@profile@') {
			print "Cannot resolve global profile when processing $file\n";
			$profile = null;
		}
		
		// if no profile name passed, see if one was defined in the 'globals'
		// section of the config for the file that we're currently reading
		if ($profile == null) {
			if (isset($config['globals']['profile'])) {
				$profile = $config['globals']['profile'];
				print "File $file: Using profile='$profile' from 'globals' section of config\n";//TODO
			} else {
				print "File $file: ignoring as no global profile set\n";
			}
		} else {
			print "File $file: using profile section '$profile'\n";
		}
		
		if (isset($config[$profile])) {
			$this->_appendMacroDefs($config[$profile]);
		}
	}
	
	function _debugOutput($string)
	{
		print "$string\n";
	}
	
	// $string is a string which may contain all sorts of stuff including macros
	function resolveString ($string, $macro_values_stack = null, $context=null) {
		if ($macro_values_stack === null) {
		if ($this->_debug > 1) $this->_debugOutput("using default stack to resolve string");
			$macro_values_stack = &$this->_macro_values;
		}
		return $this->_resolveString($string, $macro_values_stack, $context);
	}
	
	// $context is an optional context for recording errors
	function _resolveString($string, &$macro_values_stack, $context=null)
	{
		if ($this->_debug > 1) $this->_debugOutput("resolving string '$string'");

		// take a copy for debugging
		$origstring = $string;
		
		$oldstring = '';
		while ($string != $oldstring) {
			$oldstring = $string;
			$macros = $this->extractMacrosFromString($string);
			if ($this->_debug > 1) $this->_debugOutput("macros in string '$string' = ".implode(',',$macros));
			foreach ($macros as $macro) {
				$macro_contents = $this->_resolveMacro($macro, $macro_values_stack);
				if ("@$macro@" == $macro_contents) {
					// Didn't manage to resolve macro
					$hash = md5($context . $macro); // prevent duplicates
					$this->_unresolved_macros[$hash] = array('macro' => $macro, 'context' => $context);
				} else {
					$string = str_replace("@$macro@", $macro_contents, $string);
				}
			}
		}
		if ($this->_debug > 1) $this->_debugOutput("resolved string '$origstring' to '$string'");
		return $string;
	}
	
	function resolveAllMacros()
	{
		$changed = true;
		
		foreach ($this->_macro_defs as $macro) {
			$this->_macro_values[$macro->name] = $macro->value;
		}
		
		while ($changed) {
			if ($this->_debug > 1) $this->_debugOutput("resolveAllMacros: iterating");
			$changed = false;
			foreach ($this->_macro_values as $macro_name => $macro_value) {
				if ($this->_debug > 1) $this->_debugOutput("resolveAllMacros: resolving macro '$macro->name' ('$macro->value')");
				$this->_macro_values[$macro_name] = $this->resolveString($macro_value);
			
				// If macro value has changed, we will need to iterate again
				if ($this->_macro_values[$macro_name] != $macro_value) {
					$changed = true;
				}
			
				if ($this->_debug > 1) $this->_debugOutput("resolveAllMacros: resolved $macro->name to '".$this->_macro_values[$macro->name]."'");
			}
		}
		return $this->_macro_values;
	}
	
	// $macro is a SINGLE macro (not a string containing macros) something like "foo" or "hyphen:foo"
	// note "foo" not "@foo@"
	// NB no recursive replace is required here because by this stage ALL MACROS should already have been expanded
	
	function resolveMacro ($macro, $macro_values_stack = null) {
		if ($macro_values_stack === null) {
			$macro_values_stack = &$this->_macro_values;
		}
		return $this->_resolveMacro($macro, $macro_values_stack);
	}
	
	function _resolveMacro ($macro, &$macro_values_stack) {
		if ($this->_debug > 1) $this->_debugOutput("resolve single macro '$macro'");
	
		// make a copy
		$origmacro = $macro;
		
		$output = null;
	
		# separate the modifier if macro is "foo:bar" 
		if (preg_match('/([a-z]+):(.+)/',$macro,$matches)) {
			$modifier = $matches[1];
			$macro = $matches[2];
		}
		
		if (isset($macro_values_stack[$macro])) {
			if ($this->_debug) $this->_debugOutput("found macro '$macro'");
			$output = $macro_values_stack[$macro];
		} else {
			$macro_fallbacks = array(
				'vhost(\d+)_name' => 'vhost_name',
				'vhost(\d+)_interface' => 'vhost_interface',
				'vhost(\d+)_config_template' => 'vhost_config_template',
				'db(\d+)_type' => 'db_type',
				'db(\d+)_name' => 'db_name',
				'db(\d+)_host' => 'db_host',
				'db(\d+)_user' => 'db_user',
				'db(\d+)_pass' => 'db_pass',
				'db(\d+)_schema' => 'db_schema',
			 );
			$fallback_match = false;
			foreach ($macro_fallbacks as $fallback_regex => $fallback_macro) {
				if (!$fallback_match) {
					if ($this->_debug) $this->_debugOutput("trying fallback '$fallback_regex'");
					if (preg_match("/^$fallback_regex\$/", $macro, $matches)) {
						$enum_value = $matches[1];
						if ($this->_debug) $this->_debugOutput("matched fallback '$fallback_regex' - numbered '$enum_value'");
						if ($enum_value == 1) $enum_value = '';
						
						// Now fill in the special enumerated macro with the relevant number
						if (substr($fallback_macro,0,2) == 'db') {
							$special_macro = 'db_number';
						} else {
							$special_macro = 'vhost_number';
						}
						
						// Fork off a new macro values stack
						
						if ($this->_debug) $this->_debugOutput("Forking new macro values stack with $special_macro = $enum_value");
						$macro_values_special = $macro_values_stack;
						
						$macro_values_special[$special_macro] = $enum_value;
						
						$output = $this->_resolveString("@$fallback_macro@", $macro_values_special);
						$fallback_match = true;
					}
				}
			}
		}
		
		$output_unprocessed = $output;
		// if output is still null, output is just macro name
		if ($output === null) {
			$output = "@$origmacro@";
			$output_unprocessed = $output;
		} else {
			// post-process output with modifier if relevant
			if (isset($modifier)) {
				switch($modifier) {
					case 'hyphen':
						$sub = '-';
						break;
					case 'slash':
						$sub = '/';
						break;
					case 'underscore':
						$sub = '_';
						break;
					default:
						trigger_error("unknown modifier $modifier",E_USER_ERROR);
				}
				if (isset($sub)) {
					$output = str_replace($this->_resolveMacro('appref_separator', $macro_values_stack), $sub, $output);
				}
			}
		}
		
		if ($this->_debug) $this->_debugOutput("resolved single macro '$origmacro' to '$output'");
		
		$macro_values_stack[$macro] = $output_unprocessed;
		return $output;
	}
	
	
	/**
	 * Get macro name without modifiers
	 *
	 * @param string $macro_name Macro name, possibly including modifiers
	 */	
	function getBaseMacroName($macro_name)
	{
		preg_match('/([a-z]+):(.+)/', $macro_name, $matches);
		return $matches[2];
	}
	
}

class Tools_WADF_MacroDef
{
	var $name;
	var $value;
	
	function Tools_WADF_MacroDef($name, $value)
	{
		$this->name = $name;
		$this->value = $value;
	}
}

?>
